/** @file
 *
 * Datapoint-based ring buffer implementation.
 */

#uses "fwRPC/fwRPC_common.ctl"

#uses "CtrlZlib"

/**
 * @ingroup Constants
 * @{
 */
private global const string DPE_VALUE = ":_original.._value";

public global const string fwRPC_RingBuffer_DPT        = "_FwRPC_RingBuffer";
public global const string fwRPC_RingBuffer_DPE_READ   = ".read_idx";
public global const string fwRPC_RingBuffer_DPE_WRITE  = ".write_idx";
public global const string fwRPC_RingBuffer_DPE_SIZE   = ".size";
public global const string fwRPC_RingBuffer_DPE_VALUES = ".values";

public global const int fwRPC_RingBuffer_ESUCCESS   = 0;
public global const int fwRPC_RingBuffer_EERROR     = -1;
public global const int fwRPC_RingBuffer_EFULL      = -2;
public global const int fwRPC_RingBuffer_EEMPTY     = -2;
public global const int fwRPC_RingBuffer_ENODP      = -1;

public global const int fwRPC_RingBuffer_DEFAULT_SIZE = 50;
/* @} */


//------------------------------------------------------------------------------


/**
 * Initialize a ring buffer DP. Creates a new DP if the specified one does not
 *  exist.
 *
 * @param dp_rb (IN, string) DP name
 * @param size  (IN, uint)   size of the ring buffer
 *
 * @returns 0 on success
 *
 * @throws if the DP could not be created
 *         if the DP could not be written to
 */
int fwRPC_RingBuffer_init(
  string dp_rb,
  uint size = fwRPC_RingBuffer_DEFAULT_SIZE
) {
  dyn_errClass err;
  dyn_blob values;

  blob x;
  blobZero(x, 0);
  for (int i=0; i<size; i++) {
    dynAppend(values, x);
  }

  if (!dpExists(dp_rb)) {
    dpCreate(dp_rb, fwRPC_RingBuffer_DPT);
    err = getLastError();
    if (dynlen(err)) throwError(err);

    // FWCORE-3326: firsly disable last value store ...
    dpSetWait(
      dp_rb + fwRPC_RingBuffer_DPE_READ   + fwRPC_DPE_NO_VALUE_STORAGE, 1,
      dp_rb + fwRPC_RingBuffer_DPE_WRITE  + fwRPC_DPE_NO_VALUE_STORAGE, 1,
      dp_rb + fwRPC_RingBuffer_DPE_SIZE   + fwRPC_DPE_NO_VALUE_STORAGE, 1,
      dp_rb + fwRPC_RingBuffer_DPE_VALUES + fwRPC_DPE_NO_VALUE_STORAGE, 1
    );
    err = getLastError();
    if (dynlen(err)) throwError(err);
  }

  // FWCORE-3326: ...and only then (re-)set initial values
  dpSetWait(
    dp_rb + fwRPC_RingBuffer_DPE_READ   + DPE_VALUE, 1,
    dp_rb + fwRPC_RingBuffer_DPE_WRITE  + DPE_VALUE, 1,
    dp_rb + fwRPC_RingBuffer_DPE_SIZE   + DPE_VALUE, size,
    dp_rb + fwRPC_RingBuffer_DPE_VALUES + DPE_VALUE, values
  );

  err = getLastError();
  if (dynlen(err)) throwError(err);

  return fwRPC_RingBuffer_ESUCCESS;
}


/**
 * Write to a ring buffer.
 *
 * If the ring buffer is full, i.e.: read_idx == write_idx, the write will fail.
 *
 * @param dp_rb  (IN, string) ring buffer DP
 * @param value  (IN, string) value to write
 *
 * @returns  0 on success
 *          -1 if the DP doesn't exist
 *          -2 if the buffer is full
 *
 * @throws  if the DP could not be written to
 *          if the DP could not be read
 */
int fwRPC_RingBuffer_write(string dp_rb, string value) {
  int inserted;
  dyn_string values = makeDynString(value);
  int err = fwRPC_RingBuffer_writeMany(dp_rb, values, 1, inserted);

  return err;
}


/**
 * Write many values to a ringbuffer. Always returns fwRPC_RingBuffer_ESUCCESS
 * on successful insertion, even if inserted != to_insert. The value of inserted
 * should always be checked. fwRPC_RingBuffer_EFULL is not returned if only a
 * part of values is inserted.
 *
 * If to_insert is set to -1, the function will attempt to insert all the
 * elements.
 *
 * @param dp_rb     (IN, string)      ring buffer DP
 * @param values    (IN, dyn_string)  values to insert
 * @param to_insert (IN, int)         number of entries to insert. Attempts to
 *                                    insert all elements if set to -1.
 * @param inserted  (OUT, int)        number of elements actually inserted
 *
 * @returns fwRPC_RingBuffer_ESUCCESS on successful insertion
 *          fwRPC_RingBuffer_ENODP    if the DP does not exist
 *          fwRPC_RingBuffer_EFULL    if the ring buffer is full
 */
int fwRPC_RingBuffer_writeMany(
  string dp_rb,
  dyn_string values,
  int to_insert,
  int &inserted
) {
  if (!dpExists(dp_rb)) return fwRPC_RingBuffer_ENODP;

  inserted = 0;
  if (to_insert == 0) return fwRPC_RingBuffer_ESUCCESS;

  dyn_errClass err;
  uint read, write, size;
  dyn_blob raw_values;

  dpGet(
    dp_rb + fwRPC_RingBuffer_DPE_READ   + DPE_VALUE, read,
    dp_rb + fwRPC_RingBuffer_DPE_WRITE  + DPE_VALUE, write,
    dp_rb + fwRPC_RingBuffer_DPE_SIZE   + DPE_VALUE, size,
    dp_rb + fwRPC_RingBuffer_DPE_VALUES + DPE_VALUE, raw_values
  );

  err = getLastError();
  if (dynlen(err)) throwError(err);

  uint maxptr = size + read;
  if (write >= maxptr) return fwRPC_RingBuffer_EFULL;

  int num_values = dynlen(values);
  if (to_insert == -1) to_insert = num_values;

  while (
    inserted < num_values &&
    inserted < to_insert &&
    write < maxptr
  ) {
    string value = values[inserted + 1];
    blob compressed_value;
    gzip(value, compressed_value);

    raw_values[(write % size) + 1] = compressed_value;

    write++;
    inserted++;
  }

  dpSetWait(
    dp_rb + fwRPC_RingBuffer_DPE_WRITE  + DPE_VALUE, write,
    dp_rb + fwRPC_RingBuffer_DPE_VALUES + DPE_VALUE, raw_values
  );

  err = getLastError();
  if (dynlen(err)) throwError(err);

  return fwRPC_RingBuffer_ESUCCESS;
}


/**
 * Read from a ring buffer
 *
 * @param dp_rb  (IN, string)  ring buffer DP
 * @param value  (OUT, string) value read
 *
 * @returns  0 on success
 *          -1 if the DP doesn't exist
 *          -2 if there are no new values to read
 *
 * @throws  if the DP could not be read from
 *          if the DP could not be written to
 */
int fwRPC_RingBuffer_read(string dp_rb, string &value) {
  dyn_string values;
  int err = fwRPC_RingBuffer_readMany(dp_rb, values, 1);
  if (dynlen(values) == 1) value = values[1];
  return err;
}


/**
 * Read multiple entries from a ring buffer. Always returns ESUCCESS on
 * successful read, even if the number of entries read does not equal the max
 * number specified. The size of the output value should be checked.
 *
 * @param dp_rb  (IN, string)       ring buffer DP
 * @param values (OUT, dyn_string)  values read (size may != max)
 * @param max    (IN, int)          max number of values to read
 *
 * @returns fwRPC_RingBuffer_ESUCCESS on success
 *          fwRPC_RingBuffer_EEMPTY   if the buffer is empty
 *          fwRPC_RingBuffer_ENODP    if the DP does not exist
 *
 * @throws if the DP could not be read from
 *         if the DP could not be written to
 */
int fwRPC_RingBuffer_readMany(string dp_rb, dyn_string &values, int max) {
  if (!dpExists(dp_rb)) return fwRPC_RingBuffer_ENODP;

  dyn_errClass err;
  uint read, write, size;
  dyn_blob raw_values;

  dpGet(
    dp_rb + fwRPC_RingBuffer_DPE_READ   + DPE_VALUE, read,
    dp_rb + fwRPC_RingBuffer_DPE_WRITE  + DPE_VALUE, write,
    dp_rb + fwRPC_RingBuffer_DPE_SIZE   + DPE_VALUE, size,
    dp_rb + fwRPC_RingBuffer_DPE_VALUES + DPE_VALUE, raw_values
  );

  err = getLastError();
  if (dynlen(err)) throwError(err);

  //Prevent infinite looping (PSEN-2287, FWCORE-3328)
  if (read > write)
  {
    const string errMsg = "fwRPC critical error: READ index of Ring Buffer > WRITE index";
    DebugTN(__FUNCTION__, errMsg, "dp_rb", dp_rb, "max", max, "read", read, "write", write, "size", size, getStackTrace());

    // We reset the Ring Buffer, this may cause data inconsistency!
    // But it is still better then blocking the fwRPC Core manager and overloading Event Manager.
    dpSetWait(dp_rb + fwRPC_RingBuffer_DPE_READ + DPE_VALUE, 0);
    dpSetWait(dp_rb + fwRPC_RingBuffer_DPE_WRITE + DPE_VALUE, 0);
    return fwRPC_RingBuffer_EERROR;
  }

  if (read == write) return fwRPC_RingBuffer_EEMPTY;

  values = makeDynString();

  uint maxptr = read + max;
  while (read < write && read < maxptr) {
    blob compressed_value = raw_values[(read % size) + 1];
    string value;

    if (compressed_value) {
      gunzip(compressed_value, value);
    } else {
      value = "";
    }

    dynAppend(values, value);

    read++;
  }

  dpSetWait(dp_rb + fwRPC_RingBuffer_DPE_READ + DPE_VALUE, read);

  err = getLastError();
  if (dynlen(err)) throwError(err);

  return fwRPC_RingBuffer_ESUCCESS;
}
