/** @file
 *
 * This library contains helper functions for commmon PMON tasks. Currently
 *  provides facilities for CTRL manager and worker management.
 *
 * @author  Victor Rodrigues (CERN, BE-ICS-SDS)
 */

#uses "fwRPC/fwRPC_common.ctl"
#uses "fwGeneral/fwManager.ctl"

/**
 * @ingroup Constants
 * @{
 */
const int fwRPC_PMON_CONINFO_HOST = 1;
const int fwRPC_PMON_CONINFO_PORT = 2;
const int fwRPC_PMON_CONINFO_USER = 3;
const int fwRPC_PMON_CONINFO_PASS = 4;

const int fwRPC_PMON_MANINFO_INDEX = 1;
const int fwRPC_PMON_MANINFO_NUM   = 2;
const int fwRPC_PMON_MANINFO_STATE = 3;
const int fwRPC_PMON_MANINFO_CMD   = 4;

const int fwRPC_PMON_MANINFO_STATE_STOPPED  = 0;
const int fwRPC_PMON_MANINFO_STATE_INIT     = 1;
const int fwRPC_PMON_MANINFO_STATE_RUNNING  = 2;
const int fwRPC_PMON_MANINFO_STATE_BLOCKED  = 3;

const int fwRPC_PMON_STAT_STATE = 1;
const int fwRPC_PMON_STAT_NUM = 5;
const int fwRPC_PMON_LIST_CMD = 6;

const int fwRPC_PMON_STATE_WAIT_TIMESTEP = 3;
const int fwRPC_PMON_STATE_WAIT_TIMEOUT  = 10;

const string fwRPC_WORKER_SCRIPT = "fwRPC/fwRPC_worker_script.ctl";
/* @} */


//------------------------------------------------------------------------------


/**
 * Get connection info for the current system. For use with pmon functions in
 *  this module.
 *
 * @returns pmon connection info object
 *
 * @throws if pmon username/password could not be resolved
 */
dyn_mixed fwRPC_pmon_getConInfo() {
  dyn_mixed info;

  info[fwRPC_PMON_CONINFO_HOST] = fwInstallation_getHostname();
  info[fwRPC_PMON_CONINFO_PORT] = pmonPort();

  int err = fwInstallation_getPmonInfo(
    info[fwRPC_PMON_CONINFO_USER],
    info[fwRPC_PMON_CONINFO_PASS]
  );

  if (err) {
    fwException_throw(
      fwRPC_EX_PMON_AUTH,
      "Could not resolve pmon username or password."
    );
  }

  return info;
}


/**
 * Get manager list.
 *
 * @param connectionInfo (IN, dyn_mixed)      pmon connection info object
 * @param list           (OUT, dyn_dyn_mixed) manager list
 *
 * @returns nothing
 *
 * @throws if the pmon query failed
 */
void fwRPC_pmon_getManagerList(dyn_mixed connectionInfo, dyn_dyn_mixed &list) {
  string command = "MGRLIST:LIST";

  int err = _fwRPC_pmon_query(connectionInfo, command, list);
  if (err) {
    fwException_throw(
      fwRPC_EX_PMON_QUERY,
      "Could not retrieve manager list via PMON: " + err + " using conn info" +
      " host: " + connectionInfo[fwRPC_PMON_CONINFO_HOST] +
      " port: " + connectionInfo[fwRPC_PMON_CONINFO_PORT] +
      " user: " + connectionInfo[fwRPC_PMON_CONINFO_USER] +
      " pass: " + connectionInfo[fwRPC_PMON_CONINFO_PASS]
    );
  }
}


/**
 * Get manager stat list.
 *
 * @param connectionInfo (IN, dyn_mixed)      pmon connection info object
 * @param list           (OUT, dyn_dyn_mixed) manager list
 *
 * @returns nothing
 *
 * @throws if the pmon query failed
 */
void fwRPC_pmon_getManagerStati(dyn_mixed connectionInfo, dyn_dyn_mixed &list) {
  string command = "MGRLIST:STATI";

  int err = _fwRPC_pmon_query(connectionInfo, command, list);
  if (err) {
    fwException_throw(
      fwRPC_EX_PMON_QUERY,
      "Could not retrieve manager stati list via PMON"
    );
  }
}


/**
 * Get the number of managers.
 *
 * @param connectionInfo (IN, dyn_mixed) pmon connection info object
 *
 * @returns count of managers
 *
 * @throws if the pmon query failed
 */
int fwRPC_pmon_getManagerCount(dyn_mixed connectionInfo) {
  dyn_dyn_mixed managers;
  fwRPC_pmon_getManagerList(connectionInfo, managers);

  return dynlen(managers);
}


/**
 * Append a CTRL manager to run the specified script.
 *
 * @param connectionInfo (IN, dyn_mixed) mon connection info object
 * @param script         (IN, string)    file path to run when the mngr starts
 *
 * @returns index of created CTRL manager
 *
 * @throws if the pmon query failed
 */
int fwRPC_pmon_appendCtrlManager(dyn_mixed connectionInfo, string script) {
  int index = fwRPC_pmon_getManagerCount(connectionInfo);

  string command = "SINGLE_MGR:INS " + index + " WCCOActrl manual 0 0 0 " + 
                   script;

  int err = _fwRPC_pmon_command(connectionInfo, command);
  if (err) {
    fwException_throw(
      fwRPC_EX_PMON_COMMAND,
      "Could not append manager via PMON."
    );
  }

  bool success = _fwRPC_pmon_waitForState(
    connectionInfo,
    index,
    fwRPC_PMON_MANINFO_STATE_STOPPED
  );

  if (!success) DebugTN("[pmon] warning: did not observe target append for: " + script);

  return index;
}


/**
 * Start CTRL manager at index.
 *
 * @param connectionInfo (IN, dyn_mixed) pmon connection info object
 * @param index          (IN, int)       index of CTRL manager
 *
 * @returns nothing
 *
 * @throws if the pmon query failed
 */
void fwRPC_pmon_startCtrlManager(dyn_mixed connectionInfo, int index) {
  string command = "SINGLE_MGR:START " + index;

  int err = _fwRPC_pmon_command(connectionInfo, command);
  if (err) {
    fwException_throw(
      fwRPC_EX_PMON_COMMAND,
      "Could not start CTRL manager via PMON."
    );
  }

  bool success = _fwRPC_pmon_waitForState(
    connectionInfo,
    index,
    fwRPC_PMON_MANINFO_STATE_RUNNING
  );

  if (!success) DebugTN("[pmon] warning: did not observe target start for " + index);
}


/**
 * Stop CTRL manager at index.
 *
 * @param connectionInfo (IN, dyn_mixed) pmon connection info object
 * @param index          (IN, int)       index of CTRL manager
 *
 * @returns nothing
 *
 * @throws if the pmon query failed
 */
void fwRPC_pmon_stopCtrlManager(dyn_mixed connectionInfo, int index) {
  string command = "SINGLE_MGR:STOP " + index;

  int err = _fwRPC_pmon_command(connectionInfo, command);
  if (err) {
    fwException_throw(
      fwRPC_EX_PMON_COMMAND,
      "Could not stop CTRL manager via PMON."
    );
  }

  bool success = _fwRPC_pmon_waitForState(
    connectionInfo,
    index,
    fwRPC_PMON_MANINFO_STATE_STOPPED
  );

  if (!success) DebugTN("[pmon] warning: did not observe target shutdown for: " + index);
}


/**
 * Delete a CTRL manager at an index
 *
 * @param connectionInfo (IN, dyn_mixed) pmon connection info object
 * @param index          (IN, int)       index of CTRL manager
 *
 * @returns nothing
 *
 * @throws if the pmon query failed
 */
void fwRPC_pmon_deleteCtrlManager(dyn_mixed connectionInfo, int index) {
  string command = "SINGLE_MGR:DEL " + index;

  int count = fwRPC_pmon_getManagerCount(connectionInfo);

  int err = _fwRPC_pmon_command(connectionInfo, command);
  if (err) {
    fwException_throw(
      fwRPC_EX_PMON_COMMAND,
      "Could not delete CTRL manager via PMON"
    );
  }

  dyn_dyn_mixed managers;
  int timeout = fwRPC_PMON_STATE_WAIT_TIMEOUT;
  do {
    delay(fwRPC_PMON_STATE_WAIT_TIMESTEP);
    timeout -= fwRPC_PMON_STATE_WAIT_TIMESTEP;

    if (fwRPC_pmon_getManagerCount(connectionInfo) < count) return;
    
  } while (timeout > 0);

  DebugTN("[pmon] warning: did not observe target delete for: " + index);
}


/**
 * Delete the last worker. Automatically stops the CTRL manager if it's running.
 *
 * @param connectionInfo (IN, dyn_mixed) pmon connection info object
 *
 * @returns nothing
 *
 * @throws if the pmon query failed
 */
void fwRPC_pmon_deleteLastWorker(dyn_mixed connectionInfo) {
  dyn_dyn_mixed workers;

  fwRPC_pmon_getWorkerCtrlManagers(connectionInfo, workers);

  dyn_mixed lastWorker = workers[dynlen(workers)];
  int index = lastWorker[fwRPC_PMON_MANINFO_INDEX];

  string command;
  if (lastWorker[fwRPC_PMON_MANINFO_STATE] != fwRPC_PMON_MANINFO_STATE_STOPPED) {
    fwRPC_pmon_stopCtrlManager(connectionInfo, index);
  }

  fwRPC_pmon_deleteCtrlManager(connectionInfo, index);
}


/**
 * Get all CTRL managers which are set to run the worker module.
 *
 * @param connectionInfo (IN, dyn_mixed)      pmon connection info object
 * @param managers       (OUT, dyn_dyn_mixed) collection of CTRL manager objects
 *
 * @returns nothing
 *
 * @throws if the pmon query failed
 */
void fwRPC_pmon_getWorkerCtrlManagers(
  dyn_mixed connectionInfo,
  dyn_dyn_mixed &managers
) {
  string command;
  int err;

  dyn_dyn_mixed managerList;
  fwRPC_pmon_getManagerList(connectionInfo, managerList);

  dyn_dyn_mixed managerStat;
  fwRPC_pmon_getManagerStati(connectionInfo, managerStat);

  dyn_dyn_mixed managerInfo;

  int index = 1;
  for(int i=1; i<=dynlen(managerList); i++) {
    if (dynlen(managerList[i]) < fwRPC_PMON_LIST_CMD) continue;

    if (managerList[i][fwRPC_PMON_LIST_CMD] != fwRPC_WORKER_SCRIPT) continue;

    managerInfo[index][fwRPC_PMON_MANINFO_INDEX] = i - 1;
    managerInfo[index][fwRPC_PMON_MANINFO_NUM]   = (int)managerStat[i][fwRPC_PMON_STAT_NUM];
    managerInfo[index][fwRPC_PMON_MANINFO_STATE] = managerStat[i][fwRPC_PMON_STAT_STATE];
    managerInfo[index][fwRPC_PMON_MANINFO_CMD]   = managerList[i][fwRPC_PMON_LIST_CMD];

    index++;
  }

  managers = managerInfo;
}


//------------------------------------------------------------------------------


/**
 * @internal
 *
 * Execute a PMON command
 *
 * @param connectionInfo (IN, dyn_mixed) pmon connection info object
 * @param command        (IN, string)    pmon command to execute
 *
 * @returns result of pmon_command, 0 if success
 *
 * @throws nothing
 */
private int _fwRPC_pmon_command(dyn_mixed connectionInfo, string command) {
  string cmd = _fwRPC_pmon_getCredentialsString(connectionInfo) + "#" + command;

  int err = pmon_command(
    cmd,
    connectionInfo[fwRPC_PMON_CONINFO_HOST],
    connectionInfo[fwRPC_PMON_CONINFO_PORT],
    false, true
  );

  dyn_errClass error = getLastError();
  if (err || dynlen(error)) {
    fwRPC_dbg_log(
      fwRPC_DBG_EVENT_ERROR,
      "Error running pmon_command: " + error
    );
    throwError(error);
  }

  return err;
}


/**
 * @internal
 *
 * Execute a PMON query
 *
 * @param connectionInfo (IN, dyn_mixed)  pmon connection info object
 * @param query          (IN, string)     pmon query to execute
 * @param res            (OUT, dyn_dyn_mixed) result of the pmon query
 *
 * @returns result of pmon_query, 0 if success
 *
 * @throws nothing
 */
private int _fwRPC_pmon_query(dyn_mixed connectionInfo, string query, dyn_dyn_mixed &res) {
  string cmd = _fwRPC_pmon_getCredentialsString(connectionInfo) + "#" + query;

  return pmon_query(
    cmd,
    connectionInfo[fwRPC_PMON_CONINFO_HOST],
    connectionInfo[fwRPC_PMON_CONINFO_PORT],
    res,
    false, true
  );
}


/**
 * @internal
 *
 * Retrieve the credentials string required for PMON actions.
 *
 * @param connectionInfo (IN, dyn_mixed) PMON connection info
 *
 * @returns connection string
 *
 * @throws nothing
 */
private string _fwRPC_pmon_getCredentialsString(dyn_mixed connectionInfo) {
  return connectionInfo[fwRPC_PMON_CONINFO_USER] + "#" +
         connectionInfo[fwRPC_PMON_CONINFO_PASS];
}


/**
 * @intenal
 *
 * Wait for a WCCOActrl manager state change (blocking) until a specific timeout
 * has been reached.
 *
 * @param coninfo (IN, dyn_mixed) connection info object
 * @param index   (IN, int)       WCCOActrl manager index
 * @param state   (IN, int)       state to wait for
 *
 * @returns true  on success
 *          false otherwise
 *
 * @throws nothing
 */
private bool _fwRPC_pmon_waitForState(dyn_mixed coninfo, int index, int state) {
  dyn_dyn_mixed managers;
  int timeout = fwRPC_PMON_STATE_WAIT_TIMEOUT;

  do {
    delay(fwRPC_PMON_STATE_WAIT_TIMESTEP);
    timeout -= fwRPC_PMON_STATE_WAIT_TIMESTEP;

    fwRPC_pmon_getManagerStati(coninfo, managers);
    if (index >= dynlen(managers)) return false;
    if (managers[index+1][fwRPC_PMON_STAT_STATE] == state) return true;
  } while (timeout > 0);

  return false;
}


bool fwRPC_pmon_waitForWorkerStart(unsigned WorkerNum) {
  int timeout = fwRPC_PMON_STATE_WAIT_TIMEOUT;

  dyn_string err;
  dyn_string scripts = fwManager_getScriptsRunByCtrlMan("", WorkerNum, err);
  
  while(!dynContains(scripts, fwRPC_WORKER_SCRIPT) && timeout > 0) {
    scripts = fwManager_getScriptsRunByCtrlMan("", WorkerNum, err);

    if (dynlen(err)) {
      fwRPC_dbg_log(
        fwRPC_DBG_EVENT_ERROR,
        "Error calling fwManager: " + err
      );

      return false;
    }


    delay(fwRPC_PMON_STATE_WAIT_TIMESTEP);
    timeout -= fwRPC_PMON_STATE_WAIT_TIMESTEP;
  }

  return dynContains(scripts, fwRPC_WORKER_SCRIPT);
}
