/*
  ENS-18587: Change of LASER address in alarm manager
  
  This function aims at automatically migrating the Laser borker definitions with the following rules:
  sljas2:XXX -> ICEALARMS1:XXX
  sljas3:XXX -> ICEALARMS2:XXX
    
  This library is meant to be used during the migration period from WinCC OA 3.11 to 3.15
    
  11-January-2017  
  Jean-Charles Tournier
  CERN BE/ICS/SDS  
  */


/*
  Main function to migrate all the broker definition
  of all Laser managers defined in the local project.
  
  dyn_string &dsOutput: Array of string containing all the outputs written to the logs during the execution of the function
  
  @return 0 in case of success, -1 in case of failure (PVSS convention)
  */
public int unLaser_migrateBroker( dyn_string &dsOutput)
{
  //clean the output
  dynClear(dsOutput);
  
  //0 - define the mapping for the broker redefinition
  mapping mBrokerMigrationRules;
  mBrokerMigrationRules["sljas2"] = "ICEALARMS1";
  mBrokerMigrationRules["sljas3"] = "ICEALARMS2"; 
  
  //1 - search and get all Broker definition in the current system
  dyn_string dsBrokerDpe;
  dyn_string dsBrokerValue;
  
  int iNumBroker = _unLaser_getBrokerDefinition( dsBrokerDpe, dsBrokerValue, dsOutput);
  
  if ( iNumBroker < 0)
  {
    return -1;
  }
  
  if ( iNumBroker == 0) //Nothing to migrate
  {
    return 0; 
  }
  

  //2 - replacement of broker definition
  _unLaser_updateBrokerDefinition(dsBrokerValue, mBrokerMigrationRules, dsOutput);
  
  //3 - set the new broker definition
  int iErrCode = _unLaser_setBrokerDefinition(dsBrokerDpe, dsBrokerValue, dsOutput);
  
  return iErrCode;
}


/*
  Set the Broker definition, i.e. the DPE names as well as the values of the DPEs
  
  dyn_string &dsBrokerDpe (IN): list of dpe containing the broker definition
  dyn_string &dsBrokerValue (IN): list containing the broker definition
  dyn_string &dsOutput (IN/OUT): list of debug messages
  
  @return 0 in case of success, -1 in case of error
  */
private int _unLaser_setBrokerDefinition( dyn_string &dsBrokerDpe, dyn_string &dsBrokerValue, dyn_string &dsOutput)
{
  int iErr = dpSetWait( dsBrokerDpe, dsBrokerValue);
  
  dyn_errClass deErr = getLastError();

  if( iErr != 0 || dynlen(deErr))
  {
    _unLaser_debug( dsOutput, "Unable to perform dpSetWait to set the updated Laser broker definition");
    _unLaser_debug( dsOutput, (string)deErr);
    return -1;
  }  
  
  return iErr;
}


/*
  Replace each substring defined by the key of mBrokerMigrationRules in dsBrokerValue by the
  corresponding value of mBrokerMigrationRules
  
  dyn_string &dsBrokerValue (IN/OUT): input the current broker definition, output the updated broker definition
  mapping mBrokerMigrationRules(INT): the mapping definition the migration rules: the key is the string to replace and value is the replacement string
  dyn_string &dsOutput(IN/OUT): list of debug messages
  
  */
private void _unLaser_updateBrokerDefinition( dyn_string &dsBrokerValue, mapping mBrokerMigrationRules, dyn_string &dsOutput)
{
  int iNumBroker = dynlen( dsBrokerValue);
  int iNumRules = mappinglen( mBrokerMigrationRules);  
  
  for( int iBroker=1; iBroker<=iNumBroker; iBroker++)
  {
    string sOldBroker = dsBrokerValue[iBroker];
    string sNewBroker = sOldBroker;
    for( int iRule = 1; iRule <= iNumRules; iRule++)
    {
             
      strreplace( sNewBroker, 
                  mappingGetKey( mBrokerMigrationRules, iRule),
                  mappingGetValue( mBrokerMigrationRules, iRule)
                  );

    }
    
    if ( sOldBroker == sNewBroker)
    {
      _unLaser_debug( dsOutput, "Broker did not need to be updated: " + sOldBroker);
    } else
    {
      _unLaser_debug( dsOutput, "Broker updated from: " + sOldBroker + " to: " + sNewBroker);
    }
    
    dsBrokerValue[iBroker] = sNewBroker;
  }  
}

/*
  Get the Broker definition, i.e. the DPE names as well as the values of the DPEs
  
  dyn_string &dsBrokerDpe (OUT): list of dpe containing the broker definition
  dyn_string &dsBrokerValue (OUT): list containing the broker definition
  dyn_string &dsOutput (OUT): list of debug messages
  
  @return the number of broker definition found, -1 in case of error
  */
private int _unLaser_getBrokerDefinition( dyn_string &dsBrokerDpe, dyn_string &dsBrokerValue, dyn_string &dsOutput)
{
  //get the dpe of the current system
  dsBrokerDpe = dpNames("*.Settings.tcp_BrokerName", "_Un_Laser_Manager");
  
  int iNumBroker = dynlen( dsBrokerDpe);
  
  if ( iNumBroker < 1)
  {
    _unLaser_debug( dsOutput, "No LASER broker definition found. Nothing to migrate");
    return 0;
  }  
  
  //get the value of the broker definition
  int iErr = dpGet(dsBrokerDpe, dsBrokerValue);
  
  dyn_errClass deErr = getLastError();
  if ( dynlen(deErr) || iErr != 0)
  {
    _unLaser_debug( dsOutput, "Unable to perform dpGet to get the Laser broker definition");
    _unLaser_debug( dsOutput, (string)deErr);
    return -1;
  }  
  
  return iNumBroker;
}

/*
  DebugTN wrapup to have all the debug message also stored in an array
  
  dyn_string &dsOutput (OUT): array containing the list of messages
  string sMsg (INT): message to print in the log and append to the array
  */
private void _unLaser_debug(dyn_string &dsOutput, string sMsg)
{
  DebugTN(__FILE__, sMsg);
  dynAppend( dsOutput, sMsg);
}
